var LICENSE_SERVER_URL = "https://config.gravity.global/gravity/";
var SERVER_URL = LICENSE_SERVER_URL + "services/extension/config?url=" + encodeURI(location.href);
var CONFIG_LIST_URL = LICENSE_SERVER_URL + "services/extension/config/list";
var SITES_MAPPINGS_URL = "services/extension/sites/mappings";

var STARTS_WITH = 'STARTS_WITH';
var PAGE = 'PAGE';
var DOMAIN = 'DOMAIN';

var STRIPPED_WWW = 'www.';

var chatty = false;

(function init() {
    browser.storage.local.get(['gravityConsoleLevel'], function (result) {
        chatty = result && result.gravityConsoleLevel;
        chatter('Verbose logging is ON. Visit the extension options page to turn if off.');
    });
    chatter('Starting gravity extension.');
    checkBaseUrl();
})();

function checkConfigList() {
    chatter('Trying to read config list from browser local storage');
    browser.storage.local.get(["configList", "configListLifeTime", "configListCreationTime"], function (response) {
        chatter('Fetched config list, validating...');
        if (response.configList && response.configList.length && isDataValid(response.configListLifeTime, response.configListCreationTime)) {
            chatter('Config from storage is valid, check if url is suitable:', response.configList, response.configListLifeTime, response.configListCreationTime);
            if (isUrlSuitable(response.configList)) {
                chatter('Url is suitable.');
                getServerBaseUrl();
            } else {
                chatter('Check config list: Url is *NOT* suitable.');
                useOldBaseUrlForSiteMappings();
            }
        } else {
            getConfigList();
        }
    });
}

function getConfigList() {
    var xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function () {
        if (xhr.readyState === 4 && xhr.status === 200) {
            var response = JSON.parse(xhr.responseText);
            chatter('Got new config list', response);
            saveConfigListToExtensionStorage(response.configList, response.configListLifeTime);
            chatter('Config retrieved, check if url is suitable:', response.configList, response.configListLifeTime);
            if (isUrlSuitable(response.configList)) {
                chatter('Url is suitable.');
                getServerBaseUrl();
            } else {
                useOldBaseUrlForSiteMappings();
            }
        }
    };
    chatter('Trying to get new config list from', CONFIG_LIST_URL);
    xhr.open("GET", CONFIG_LIST_URL, true);
    xhr.send();
}

function isUrlSuitable(list) {
    var url = getStrippedUrl();
    var urlDomain = getUrlDomain();

    var hashedUrl;
    var bitArray;
    var result = false;
    for (var i = 0; i < list.length; i++) {
        var hashType = list[i].hashType;
        switch (true) {
            case hashType.indexOf(STARTS_WITH) > -1:
            {
                var splitted = hashType.split('_');
                var urlLength = Number(splitted[splitted.length - 1]);
                bitArray = sjcl.hash.sha256.hash(url.substring(0, urlLength));
                hashedUrl = sjcl.codec.hex.fromBits(bitArray);
                break;
            }
            case hashType.indexOf(PAGE) > -1:
            {
                bitArray = sjcl.hash.sha256.hash(url);
                hashedUrl = sjcl.codec.hex.fromBits(bitArray);
                break;
            }
            case hashType.indexOf(DOMAIN) > -1:
            {
                bitArray = sjcl.hash.sha256.hash(urlDomain);
                hashedUrl = sjcl.codec.hex.fromBits(bitArray);
                break;
            }
        }

        result = hashedUrl === list[i].hash;
        if (result) {
            chatter('Found match, since url\'s are hashed, no match info can be given.');
            return result;
        }
    }
    return result;
}

function useOldBaseUrlForSiteMappings() {
    browser.storage.local.get(["baseUrl"], function (response) {
        if (response.baseUrl) {
            chatter('#################### Cached base url was used. Mapping list will be checked');
            checkMappingList(response.baseUrl);
        }
    });
}

function checkBaseUrl() {
    browser.storage.local.get(["baseUrl", "baseUrlLifeTime", "baseUrlCreationTime"], function (response) {
        if (response.baseUrl && isDataValid(response.baseUrlLifeTime, response.baseUrlCreationTime)) {
            checkMappingList(response.baseUrl);
        } else {
            checkConfigList()
        }
    });
}

function getServerBaseUrl() {
    var xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function () {
        if (xhr.readyState === 4 && xhr.status === 200) {
            var response = JSON.parse(xhr.responseText);
            chatter('Got new base url', response);
            var baseUrl = response.serverUrl;
            var baseUrlLifeTime = response.serverUrlLifeTime;
            if (baseUrl && baseUrlLifeTime) {
                chatter('Base URL is valid:', baseUrl, baseUrlLifeTime);
                saveBaseUrlInfoToExtensionStorage(baseUrl, baseUrlLifeTime);
                checkMappingList(baseUrl)
            } else {
                useOldBaseUrlForSiteMappings();
            }
        }
    };
    chatter('Trying to get new base url from', SERVER_URL);
    xhr.open("GET", SERVER_URL, true);
    xhr.send();
}

function checkMappingList(baseUrl) {
    chatter('Trying to read mapping list from browser local storage');
    browser.storage.local.get(["mappingList", "sitesMappingsLifeTime", "sitesMappingsCreationTime"], function (response) {
        chatter('Fetched mapping list, validating...');
        if (response.mappingList && response.mappingList.length && isDataValid(response.sitesMappingsLifeTime, response.sitesMappingsCreationTime)) {
            chatter('Sites mappings from storage is valid, check if url is suitable:', response.mappingList, response.sitesMappingsLifeTime, response.sitesMappingsCreationTime);
            if (isUrlSuitable(response.mappingList)) {
                chatter('Url is suitable.');
                loadInjectScripts(baseUrl);
            } else {
                chatter('Url is *NOT* suitable.');
            }
        } else {
            getSitesMappings(baseUrl);
        }
    });
}

function getSitesMappings(baseUrl) {
    var xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function () {
        if (xhr.readyState === 4 && xhr.status === 200) {
            var response = JSON.parse(xhr.responseText);
            chatter('Got new sites mappings', response);
            saveMappingListToExtensionStorage(response.mappingList, response.sitesMappingsLifeTime);
            if (isUrlSuitable(response.mappingList)) {
                chatter('Url is suitable.');
                loadInjectScripts(baseUrl);
            }
        }
    };
    xhr.open("GET", baseUrl + SITES_MAPPINGS_URL, true);
    xhr.send();
}

function loadInjectScripts(baseUrl) {
    chatter('Loading inject scripts from', baseUrl);
    var scriptUrl = baseUrl + "inject/";
    var browserExtensionInfo = "Firefox extension - " + browser.runtime.getManifest().version;
    var script = document.createElement('script');
    script.onload = function () {
        var startScript = document.createElement('script');
        var srcBody = document.createTextNode("pi.gravity.inject.setup('" + baseUrl + "', '" + scriptUrl + "', '" + browserExtensionInfo +"')");
        startScript.appendChild(srcBody);
        document.head.appendChild(startScript);
    };
    script.src = scriptUrl + "js/pi.gravity.inject.js";
    document.head.appendChild(script);
}

function saveBaseUrlInfoToExtensionStorage(baseUrl, baseUrlLifeTime) {
    var container = {
        "baseUrl": baseUrl,
        "baseUrlLifeTime": baseUrlLifeTime,
        "baseUrlCreationTime": new Date().getTime()
    };
    browser.storage.local.set(container, function () {
        chatter('Base URL info saved to browser.storage.local', container);
    });

}

function saveConfigListToExtensionStorage(configList, configListLifeTime) {
    var container = {
        "configList": configList,
        "configListLifeTime": configListLifeTime,
        "configListCreationTime": new Date().getTime()
    };
    browser.storage.local.set(container, function () {
        chatter('Site config saved to browser.storage.local', container);
    });
}

function saveMappingListToExtensionStorage(mappingList, sitesMappingsLifeTime) {
    var container = {
        "mappingList": mappingList,
        "sitesMappingsLifeTime": sitesMappingsLifeTime,
        "sitesMappingsCreationTime": new Date().getTime()
    };
    browser.storage.local.set(container, function () {
        chatter('Sites mappings saved to browser.storage.local', container);
    });
}

function getUrlDomain() {
    var result = location.hostname;
    if (result.substring(0, 4) === STRIPPED_WWW) {
        result = result.substring(STRIPPED_WWW.length);
    }
    return result;
}

function getStrippedUrl() {
    var result;
    result = location.hostname + location.pathname;
    if (result.substring(0, 4) === STRIPPED_WWW) {
        result = result.substring(STRIPPED_WWW.length);
    }
    return result;
}

function isDataValid(baseUrlLifeTime, baseUrlCreationTime) {
    var currentTimestamp = new Date().getTime();
    var dataValid = currentTimestamp - baseUrlCreationTime < baseUrlLifeTime;
    chatter(dataValid ? 'Data is valid/not expired' : 'Data is *EXPIRED*');
    return dataValid;
}

function chatter() {
    if (chatty) {
        console.log.apply(this, arguments);
    }
}